#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

// Device function for scaling
__device__ float scale_value(float val, float scale) {
    return val * scale;
}

// Device function for tanh
__device__ float apply_tanh(float val) {
    return tanh(val);
}

// Device function for bias multiplication
__device__ float multiply_bias(float val, float bias) {
    return val * bias;
}

// Device function for sigmoid
__device__ float apply_sigmoid(float val) {
    return 1.0f / (1.0f + exp(-val));
}

// CUDA kernel using manual loop unrolling for better performance
__global__ void conv3d_unroll_kernel(
    const float* __restrict__ output,
    const float* __restrict__ scaling_factor,
    const float* __restrict__ bias,
    float* __restrict__ result,
    const int batch_size,
    const int out_channels,
    const int depth,
    const int height, 
    const int width) {
    
    extern __shared__ char smem[];
    float* s_scaling = reinterpret_cast<float*>(smem);
    float* s_bias = s_scaling + out_channels;

    if (threadIdx.x < out_channels) {
        s_scaling[threadIdx.x] = scaling_factor[threadIdx.x];
        s_bias[threadIdx.x] = bias[threadIdx.x];
    }
    __syncthreads();

    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int total_elements = batch_size * out_channels * depth * height * width;

    if (idx < total_elements) {
        const int c = (idx / (width * height * depth)) % out_channels;

        float val = output[idx];
        val = scale_value(val, s_scaling[c]);
        val = apply_tanh(val);
        val = multiply_bias(val, s_bias[c]);
        val = apply_sigmoid(val);

        result[idx] = val;
    }
}

// Forward function that performs conv3d followed by scaling, tanh, bias multiplication and sigmoid
// It launches the optimized kernel with loop unrolling

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor conv_weight,
    torch::Tensor conv_bias,
    torch::Tensor scaling_factor,
    torch::Tensor bias) {

    auto conv_out = torch::conv3d(x, conv_weight, conv_bias);
    
    const int batch_size = conv_out.size(0);
    const int out_channels = conv_out.size(1);
    const int depth = conv_out.size(2);
    const int height = conv_out.size(3);
    const int width = conv_out.size(4);

    auto result = torch::empty_like(conv_out);

    const int threads = 256;
    const int total_elements = batch_size * out_channels * depth * height * width;
    const int blocks = (total_elements + threads - 1) / threads;

    conv3d_unroll_kernel<<<blocks, threads, 2 * out_channels * sizeof(float)>>>(
        conv_out.data_ptr<float>(),
        scaling_factor.data_ptr<float>(),
        bias.data_ptr<float>(),
        result.data_ptr<float>(),
        batch_size,
        out_channels, 
        depth,
        height,
        width
    );

    return result;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Conv3d scale tanh bias sigmoid forward with unrolling");
}